#ifndef SB_GEO_MIPMAP_PRIMITIVES_H
#define SB_GEO_MIPMAP_PRIMITIVES_H

///////////////////////////////////////////////////////////////////////////////
//  SoTerrain
///////////////////////////////////////////////////////////////////////////////
/// Zkladn datov prvky algoritmu Geo Mip-mapping.
/// \file SbGeoMipmapPrimitives.h
/// \author Radek Barto - xbarto33
/// \date 30.01.2006
///
/// Zde jsou definovny zkladn datov typy algoritmu Geo Mip-Mapping.
/// Kvadrantov strom dladic ve td ::SbGeoMipmapTileTree rekurzivn
/// rozdluje vkovou mapu ternu na dladice tdy SbGeoMipmapTile. V jeho
/// hornch patrech jsou dladice pouze virtuln, tj. ty, kter neobsahuj
/// vlastn geometrii, ale jenom sv potomky. Dladice na nejni rovni
/// stromu obsahuj pole jednotlivch rovn detail dladice ve td
/// ::SbGeoMipmapTileLevel, jen u geometrii zahrnuj.
///////////////////////////////////////////////////////////////////////////////
//
// THIS SOFTWARE IS NOT COPYRIGHTED
//
// This source code is offered for use in the public domain.
//  You may use, modify or distribute it freely.
//
// This source code is distributed in the hope that it will be useful but
// WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
// DISCLAIMED.  This includes but is not limited to warranties of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// If you find the source code useful, authors will kindly welcome
// if you give them credit and keep their names with their source code.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef NULL
#define NULL 0
#endif

// Coin includy
#include <Inventor/SbLinear.h>
#include <Inventor/SbBox.h>

// lokalni includy
#include <debug.h>
#include <utils.h>

/** rove detail dladice algoritmu Geo Mip-Mapping.
Tato tda obsahuje indexy vrchol geometrie rovn detail jedn dladice.
Dle je zde statick st chybov metriky platn pro tuto rove, kter se
pedpota v dob natn vstupn vkov mapy. */
struct SbGeoMipmapTileLevel
{
  public:
    /* Metody */
    /** Konstruktor.
    Inicializuje instanci tdy ::SbGeoMipmapTileLevel, aby obsahovala
    pole vrchol rovn detail dladice \p vertices s celkovou statickou
    st chybov metriky \p error.
    \param error Chyba statick sti chybov metriky.
    \param vertices Pole vrchol geometrie rovn detail dladice. */
    SbGeoMipmapTileLevel(float error = 0.0f, int * vertices = NULL);
    /** Destruktor.
    Uvoln pole vrchol rovn detail dladice a deinicalizuje instanci
    tdy SbGeoMipmapTileLevel. */
    ~SbGeoMipmapTileLevel();
    /* Datove polozky */
    /// Statick st chybov metriky rovn detail dladice.
    float error;
    /// Vrcholy geometrie rovn detail dladice.
    int * vertices;
};

/** Dladice algoritmu Geo Mip-Mapping.
Z tchto dladic se vytvo kvadrantov strom dladic slouc pro vpoet
oezvn pohledovm tlesem. Pokud je ukazatel na rovn datail dladice
roven \p NULL, jedn se o virtuln dladici v hornch rovnch stromu. */
struct SbGeoMipmapTile
{
  public:
    /* Metody */
    /** Konstruktor.
    Vytvo instanci tdy ::SbGeoMipmapTile a inicializuje jeho datov
    prvky na hodnoty \p levels, \p left, \p right, \p top, \p bottom, \p level,
    \p bounds a \p center.
    \param levels Ukazatel na pole rovn detail dladice.
    \param left Ukazatel na levho souseda dladice.
    \param right Ukazatel na pravho souseda dladice.
    \param top Ukazatel na hornho souseda dladice.
    \param bottom Ukazatel na dolnho souseda dladice.
    \param level slo rovn detail, kter se m vykreslit.
    \param bounds Ohranien vrchol dladice.
    \param center Sted ohranien vrchol dladice. */
    SbGeoMipmapTile(SbGeoMipmapTileLevel * levels = NULL,
      SbGeoMipmapTile * left = NULL, SbGeoMipmapTile * right = NULL,
      SbGeoMipmapTile * top = NULL, SbGeoMipmapTile * bottom = NULL,
      int level = LEVEL_NONE, SbBox3f bounds = SbBox3f(),
      SbVec3f center = SbVec3f());
    /** Destruktor.
    Zru instanci tdy ::SbGeoMipmapTile a uvoln pole rovni detail
    dladice. */
    ~SbGeoMipmapTile();
    /* Datove polozky */
    /// Ukazatel na pole rovn detail dladice.
    SbGeoMipmapTileLevel * levels;
    /// Ukazatel na levho souseda dladice.
    SbGeoMipmapTile * left;
    /// Ukazatel na pravho souseda dladice.
    SbGeoMipmapTile * right;
    /// Ukazatel na hornho souseda dladice.
    SbGeoMipmapTile * top;
    /// Ukazatel na dolnho souseda dladice.
    SbGeoMipmapTile * bottom;
    /// slo rovn detail, kter se m vykreslit.
    int level;
    /// Ohranien vrchol dladice.
    SbBox3f bounds;
    /// Sted ohranien vrchol dladice.
    SbVec3f center;
    /* Konstanty. */
    /// slo rovn detail dladice, kter se nem vykreslovat vbec.
    static const int LEVEL_NONE;
};

/** Strom dladic algoritmu Geo Mip-Mapping.
Kvadrantov strom dladic se vytvo v prbhu natn vkov mapy a napln
se dladicemi tdy ::SbGeoMipmapTile obsahujcmi jednotliv rovn detail
dladice tdy ::SbGeoMipmapTileLevel. */
struct SbGeoMipmapTileTree
{
  public:
    /* Metody. */
    /** Konstruktor.
    Vytvo kvadrantov strom dladic algoritmu Geo Mip-Mapping tdy
    ::SbGeoMipmapTileTree obsahujcho dladice o velikosti strany \p tile_size
    uspodanch do tverce o stran \p tile_count dladic.
    \param tile_count Poet dladic na stranu vkov mapy.
    \param tile_size Velikost strany jedn dladice. */
    SbGeoMipmapTileTree(int tile_count, int tile_size);
    /** Destruktor.
    Zru instanci tdy ::SbGeoMipmapTileTree a uvoln pole velikost rovn
    detail dladic i vechny dladice ve strom. */
    ~SbGeoMipmapTileTree();
    /* Datove polozky. */
    /// Poet dladic na stranu vstupn vkov mapy.
    int tile_count;
    /// Velikost strany jedn dladice.
    int tile_size;
    /// Vka kvadrantovho stromu dladic.
    int tree_size;
    /// Index v poli dladic, kde zan nejni patro stromu.
    int bottom_start;
    /// Poet rovn detail kad dladice v nejnim patru stromu.
    int level_count;
    /// Pole velikost stran pol vrchol rovn kad dladice.
    int * level_sizes;
    /// Pole jednotlivch dladic kvadrantovho stromu dladic.
    SbGeoMipmapTile * tiles;
  private:
    /** Koprovac konstruktor.
    Zprivatizovn, aby se zabrnilo vytven kopi kvadrantovho stromu dladic.
    \param old_tree Pvodn kvadrantov strom dladic. */
    SbGeoMipmapTileTree(const SbGeoMipmapTileTree & old_tree);
};

#endif
