#ifndef SO_SIMPLE_GEO_MIPMAP_TERRAIN_H
#define SO_SIMPLE_GEO_MIPMAP_TERRAIN_H

///////////////////////////////////////////////////////////////////////////////
//  SoTerrain
///////////////////////////////////////////////////////////////////////////////
/// Tern vykreslovan algoritmem Geo Mip-Mapping.
/// \file SoSimpleGeoMipmapTerrain.h
/// \author Radek Barto - xbarto33
/// \date 29.01.2006
///
/// Uzel grafu scny reprezentujc tern vykreslovan algoritmem Geo
/// Mip-Mapping. Pro pouit je teba uzlu pedadit uzel s koordinty typu
/// SoCordinate3 obsahujc vkovou mapu ternu a nastavit jej rozmry a
/// rozmry dladic generovanch z tto vkov mapy Velikost vkov mapy mus
/// bt tvercov o velikosti strany 2^n + 1, kde n je cel kladn slo.
///////////////////////////////////////////////////////////////////////////////
//
// THIS SOFTWARE IS NOT COPYRIGHTED
//
// This source code is offered for use in the public domain.
//  You may use, modify or distribute it freely.
//
// This source code is distributed in the hope that it will be useful but
// WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
// DISCLAIMED.  This includes but is not limited to warranties of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// If you find the source code useful, authors will kindly welcome
// if you give them credit and keep their names with their source code.
//
///////////////////////////////////////////////////////////////////////////////

// OpenInventor includy
#include <Inventor/SbBasic.h>
#include <Inventor/SbLinear.h>
#include <Inventor/SoPrimitiveVertex.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/nodes/SoShape.h>
#include <Inventor/elements/SoMaterialBindingElement.h>
#include <Inventor/elements/SoCoordinateElement.h>
#include <Inventor/elements/SoTextureCoordinateElement.h>
#include <Inventor/elements/SoNormalElement.h>
#include <Inventor/elements/SoNormalBindingElement.h>
#include <Inventor/elements/SoViewVolumeElement.h>
#include <Inventor/elements/SoViewportRegionElement.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/bundles/SoMaterialBundle.h>
#include <Inventor/bundles/SoTextureCoordinateBundle.h>

// OpenGL includy
#if defined(__WIN32__) || defined(_WIN32)
  #include <windows.h>
#endif
#include <GL/gl.h>

// ostatni includy
#include <assert.h>

// lokalni includy
#include <geomipmapping/SbGeoMipmapPrimitives.h>
#include <profiler/PrProfiler.h>
#include <debug.h>

/** Tern vykreslovan algoritmem Geo Mip-Mapping.
Uzel grafu scny reprezentujc tern vykreslovan algoritmem Geo
Mip-Mapping. Pro pouit je teba uzlu pedadit uzel s koordinty typu
\p SoCordinate3 obsahujc vkovou mapu ternu a nastavit jej rozmry a
rozmry dladic generovanch z tto vkov mapy. Velikost vkov mapy mus
bt tvercov o velikosti strany 2^n + 1, kde n je cel kladn slo. */
class SoSimpleGeoMipmapTerrain : public SoShape
{
  SO_NODE_HEADER(SoSimpleGeoMipmapTerrain);
  public:
    /* Metody */
    /** Run-time inicializace tdy.
    Tuto metodu je teba zavolat ped vytvoenm jakkoliv instance tdy
    ::SoSimpleGeoMipmapTerrain.*/
    static void initClass();
    /** Konstruktor.
    Vytvo instanci tdy ::SoSimpleGeoMipmapTerrain. */
    SoSimpleGeoMipmapTerrain();
    /* Pole. */
    /// Velikost strany vstupn vkov mapy.
    SoSFInt32 mapSize;
    /// Velikost strany dladic.
    SoSFInt32 tileSize;
    /// Chyba zobrazen v pixelech.
    SoSFInt32 pixelError;
    /// Pznak oezvn pohledovm tlesem.
    SoSFBool frustrumCulling;
    /// Pznak "zmrazen" vykreslovn ternu.
    SoSFBool freeze;
  protected:
    /* Metody */
    /** Vykreslen ternu.
    Pi prvnm sputn vybuduje ze vstupn vkov mapy kvadrantov strom
    dladic. Nsledn a i pi dalch sputnch zaktualizuje vybranou rove
    detail dladic ve stromu podle pozice a natoen kamery a vybran rovn
    vykresl.
    \param action Objekt nesouc informace o grafu scny. */
    virtual void GLRender(SoGLRenderAction * action);
    /** Vytvoen vech trojhelnk vkov mapy.
    Ze vstupn vkov mapy vytvo seznam trojhelnk na nejvy rovni
    detail pro ely detekce koliz a podobn.
    \param action Objekt nesouc informace o grafu scny. */
    virtual void generatePrimitives(SoAction * action);
    /** Vpoet ohranien ternu.
    Vypote rozmry a sted kvdru, kter ohraniuje cel tern.
    \param action Objekt nesouc informace o grafu scny.
    \param box Vypoten kvdr ohraniujc tern.
    \param center Vypoten sted kvdru ohraniujcho tern. */
    virtual void computeBBox(SoAction * action, SbBox3f & box,
      SbVec3f & center);
    /** Inicializace kvadrantovho stromu dladic.
    Rekurzivn inicializuje strom dladic podle zadan vkov mapy a pot
    statickou st chybov metriky. Parametr \p index slou jako index dladice,
    kter tvo koen inicializovanho podstromu stromu dladic. Parametr
    \p coord_box slou k pedvn rozsahu index bod vstupn vkov mapy.
    \param index Index koene inicializovanho podstromu stromu dladic.
    \param coord_box Rozsah index bod vstupn vkov mapy.
    \return */
    SbBox3f initTree(const int index, SbBox2s coord_box);
    /** Inicializace jedn dladice.
    Inicializuje jednu dladici \p tile kvadrantovho stromu dladic vetn
    jejch rovn detail body vstupn vkov mapy v rozsahu index
    \p coord_box.
    \param tile Dladice, kter se m inicializovat.
    \param coord_box Rozsah index vstupn vkov mapy na zem dladice. */
    inline void initTile(SbGeoMipmapTile & tile, SbBox2s coord_box);
    /** Inicializace rovn detail dladice.
    Inicializuje rove detail \p level o velikosti strany \p level_size
    podle ji zinicializovan vy rovn detail dladice \p parent.
    \param level Inicializovan rove detail dladice.
    \param parent Ji inicializovan vy rove detail dladice.
    \param level_size Velikost strany rovn detail dladice. */
    inline void initLevel(SbGeoMipmapTileLevel & level,
      SbGeoMipmapTileLevel & parent, const int level_size);
    /** Pepotn kvadrantovho stromu dladic.
    Podle vpotu dynamick sti chybov metriky a podle pozice pohledovho
    tlesa vybere u kad dladice stromu pslunou rove detail. Toto
    provd rekurzivn od indexu podstromu \p index. Parametr \p render_parent
    slou k propagaci informace zda se se vykresluje rodiovsk dladice koene
    podstromu.
    \param index Index koene podstromu stromu dladic.
    \param render_parent Pznak toho, e se rodi koene podstromu vykresluje. */
    void recomputeTree(const int index, const SbBool render_parent);
    /** Vykreslen kvadrantovho stromu dladic.
    Rekurzivn projde kvadrantovm stromem dladic podle indexu koene \p index
    a vykresl jeho dladice s vybranou rovn detail.
    \param action Objekt nesouc informace o grafu scny.
    \param index Index koene podstromu stromu dladic. */
    void renderTree(SoAction * action, const int index);
    /** Vbr rovn detail dladice.
    Podle chybov metriky vybere sprvnou rove detail dladice, kter se m
    vykreslit.
    \param tile Dladice, u n vbr probh.
    \return Index vybran rovn detail. */
    inline int pickLevel(const SbGeoMipmapTile & tile) const;
    /* Callbacky. */
    /** Callback zmny pole \p mapSize.
    Pi zmn hodnoty pole \p mapSize nastav jeho intern reprezentaci novou
    hodnotu.
    \param instance Ukazatel na instanci tdy SoSimpleGeoMipmapTerrain.
    \param sensor Senzor, kter callback vyvolal */
    static void mapSizeChangedCB(void * instance, SoSensor * sensor);
    /** Callback zmny pole \p tileSize.
    Pi zmn hodnoty pole \p tileSize nastav jeho intern reprezentaci novou
    hodnotu.
    \param instance Ukazatel na instanci tdy SoSimpleGeoMipmapTerrain.
    \param sensor Senzor, kter callback vyvolal */
    static void tileSizeChangedCB(void * instance, SoSensor * sensor);
    /** Callback zmny pole \p pixelError.
    Pi zmn hodnoty pole \p pixelError nastav jeho intern reprezentaci
    novou hodnotu.
    \param instance Ukazatel na instanci tdy SoSimpleGeoMipmapTerrain.
    \param sensor Senzor, kter callback vyvolal */
    static void pixelErrorChangedCB(void * instance, SoSensor * sensor);
    /** Callback zmny pole \p frustrumCulling.
    Pi zmn hodnoty pole \p frustrumCulling nastav jeho intern reprezentaci
    novou hodnotu.
    \param instance Ukazatel na instanci tdy SoSimpleGeoMipmapTerrain.
    \param sensor Senzor, kter callback vyvolal */
    static void frustrumCullingChangedCB(void * instance, SoSensor * sensor);
    /** Callback zmny pole \p freeze.
    Pi zmn hodnoty pole \p freeze nastav jeho intern reprezentaci novou
    hodnotu.
    \param instance Ukazatel na instanci tdy SoSimpleGeoMipmapTerrain.
    \param sensor Senzor, kter callback vyvolal */
    static void freezeChangedCB(void * instance, SoSensor * sensor);
    /* Zkratky elementu. */
    /// Body vykov mapy.
    const SbVec3f * coords;
    /// Texturov souadnice pro body vkov mapy.
    const SbVec2f * texture_coords;
    /// Normly.
    const SbVec3f * normals;
    /// Pohledov tleso.
    const SbViewVolume * view_volume;
    /// Vykreslovac okno.
    const SbViewportRegion * viewport_region;
    /* Datove polozky. */
    /// Kvadrantov strom dladic.
    SbGeoMipmapTileTree * tile_tree;
    /// Konstanta pro vpoet dynamick sti chybov metriky.
    float distance_const;
    /// Pznak pouit textury.
    SbBool is_texture;
    /// Pznak pouit morml.
    SbBool is_normals;
    /* Interni pole. */
    /// Velikost strany vstupn vkov mapy.
    int map_size;
    /// Velikost strany dladic.
    int tile_size;
    /// Chyba zobrazen v pixelech.
    int pixel_error;
    /// Pznak oezvn pohledovm tlesem.
    SbBool is_frustrum_culling;
    /// Pznak "zmrazen" vykreslovn ternu.
    SbBool is_freeze;
    /* Sensory. */
    /// Senzor pole \p mapSize.
    SoFieldSensor * map_size_sensor;
    /// Senzor pole \p tileSize.
    SoFieldSensor * tile_size_sensor;
    /// Senzor pole \p pixelError.
    SoFieldSensor * pixel_error_sensor;
    /// Senzor pole \p frustrumCulling.
    SoFieldSensor * frustrum_culling_sensor;
    /// Senzor pole \p freeze.
    SoFieldSensor * freeze_sensor;
    /* Konstanty. */
    /// Vchoz hodnota chyby zobrazen v pixelech.
    static const int DEFAULT_PIXEL_ERROR;
  private:
    /* Metody */
    /** Destruktor.
    Zprivatizovn, aby se zabrnilo ruen uzlu, nebo Inventor uvoluje pam
    ve vlastn reii. */
    virtual ~SoSimpleGeoMipmapTerrain();
};

#endif
