#ifndef SB_ROAM_PRIMITIVES_H
#define SB_ROAM_PRIMITIVES_H

///////////////////////////////////////////////////////////////////////////////
//  SoTerrain
///////////////////////////////////////////////////////////////////////////////
/// Zkladn datov prvky algoritmu ROAM.
/// \file SbROAMPrimitives.h
/// \author Radek Barto - xbarto33
/// \date 16.09.2005
///
/// Algoritmus ROAM ve sv statick sti vyuv jako zkladn datovou
/// strukturu binrn strom trojhelnk. Pro tyto trojhelnky je zde
/// definovna tda ::SbROAMTriangle. V dynamick sti jsou pak poteba
/// dv prioritn fronty. Prvn z nich je prioritn fronta trojhelnk na
/// rozdlen. Obsahuje trojhelnky tdy ::SbROAMSplitQueueTriangle.
/// Druhou pak prioritn fronta trojhelnk na spojen. Ta obsahuje diamondy
/// (tveice trojhelnk) tdy ::SbROAMMergeQueueDiamond.
///////////////////////////////////////////////////////////////////////////////
//
// THIS SOFTWARE IS NOT COPYRIGHTED
//
// This source code is offered for use in the public domain.
//  You may use, modify or distribute it freely.
//
// This source code is distributed in the hope that it will be useful but
// WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
// DISCLAIMED.  This includes but is not limited to warranties of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// If you find the source code useful, authors will kindly welcome
// if you give them credit and keep their names with their source code.
//
///////////////////////////////////////////////////////////////////////////////

// OpenInventor includy
#include <Inventor/SbVec3f.h>
#include <Inventor/SbHeap.h>

// lokalni includy
#include <debug.h>

/** Trojhelnk binrnho stromu trojhelnk algoritmu ROAM.
Zkladn datovou strukturou algoritmu ROAM je binrn strom trojhelnk, kter
obsahuje prv tyto struktury reprezentujc jeden trojhelnk
v triangulaci. */
struct SbROAMTriangle
{
  public:
    /* Metody. */
    /** Konstruktor.
    Vytvo instanci tdy ::SbROAMTriangle a inicializuje jeho atributy.
    Parametry \p first, \p second a \p apex jsou indexy vrchol trojhelnku
    v seznamu koordint a parametr \p level je slo rovn trojhelnku
    v binrnm stromu trojhelnk.
    \param first Prvn (lev) vrchol pepony trojhelnku.
    \param second Druh (prav) vrchol pepony trojhelnku.
    \param apex Prav (nad peponou) vrchol trojhelnku.
    \param level rove, na kter se trojhelnk nachz v binrnm stromu
       trojhelnk. */
    SbROAMTriangle(int first = -1, int second = -1, int apex = -1,
      int level = 0);
    /* Datove polozky. */
    /// Lev vrchol pepony.
    int first;
    /// Prav vrchol pepony.
    int second;
    /// Vrchol nad peponou.
    int apex;
    /// rove trojhelnku ve stromu.
    int level;
    /// Chyba nezavisl na bodu pohledu.
    float error;
    /// Polomr kuloplochy ohraniujc trojhelnk a jeho potomky.
    float radius;
};

/// Trojhelnky s touto prioritou jsou zobrazeny vdy.
const float PRIORITY_MAX = 1e38f;
/// Trojhelnky s touto prioritou se nezobrazuj (teoreticky).
const float PRIORITY_MIN = 0.0f;

struct SbROAMMergeQueueDiamond;
class SbROAMSplitQueue;

/** Reprezentace trojhelnk algoritmu ROAM ve front.
Tda reprezentuje jeden trojhelnk v triangulaci. Roziuje trojhelnk
z binrnho stromu trojhelnk o informace o sousedech a priorit
ve front. */
struct SbROAMSplitQueueTriangle
{
  public:
    /* Metody. */
    /** Konstruktor.
    Vytvo instanci tdy trojhelnku \p triangle pro prioritn frontu
    trojhelnk na rozdlen a inicializuje jeho ukazatele na sousedy
    \p left, \p right a \p base. Prioritu nastav na hodnotu \p priority.
    \param triangle Trojhelnk v binrnm stromu trojhelnk.
    \param priority Priorita trojhelnku ve front.
    \param left Lev soused trojhelnku.
    \param right Prav soused trojhelnku.
    \param base Zkladnov soused trojhelnku.
    \param diamond Diamant, do kterho trojhelnk pat. */
    SbROAMSplitQueueTriangle(SbROAMTriangle * triangle,
      float priority = PRIORITY_MIN,
      SbROAMSplitQueueTriangle * left = NULL,
      SbROAMSplitQueueTriangle * right = NULL,
      SbROAMSplitQueueTriangle * base = NULL,
      SbROAMMergeQueueDiamond * diamond = NULL);
    /** Zskn priority trojhelnku ve front.
    Vrt hodnotu priority trojhelnku ve front trojhelnk na rozdlen.
    \return Priorita trojhelnku ve front trojhelnk na rozdlen. */
    inline float getPriority() const;
    /** Nastaven priority trojhelnku ve front.
    Nastav prioritu trojhelnku ve fronte \p split_queue na novou hodnotu
    \p priority.
    \param split_queue Fronta, v n se trojhelnk nachz.
    \param priority Nov priorita trojhelnku. */
    inline void setPriority(SbROAMSplitQueue * split_queue, const float
      priority);
    /** Nastaven priority trojhelnku.
    Nastav prioritu trojhelnku na novou hodnotu \p priority, ale
    neaktualizuje jeho pozici ve front.
    \param priority Nov priorita trojhelnku. */
    inline void setPriority(const float priority);
    /* Datove polozky. */
    /// Trojhelnk v binrnm stromu trojhelnk.
    SbROAMTriangle * triangle;
    /// Soused nad levou odvsnou.
    SbROAMSplitQueueTriangle * left;
    /// Soused nad pravou odvsnou.
    SbROAMSplitQueueTriangle * right;
    /// Soused pod peponou.
    SbROAMSplitQueueTriangle * base;
    /// Diamant, pod kter trojhelnk pat.
    SbROAMMergeQueueDiamond * diamond;
  private:
    /// Priorita trojhelnku ve front trojhelnk na rozdlen.
    float priority;
    /// Index trojhelnku ve front trojhelnk na rozdlen.
    int index;
  /** Prioritn fronta trojhelnk na rozdlen me pistupovat k privtnm
  atributm trojhelnku. */
  friend class SbROAMSplitQueue;
};

class SbROAMMergeQueue;

/** Uzel pedstavujc diamant algoritmu ROAM ve front.
Tda pedstavuje diamant, tj. tyi trojhelniky, kter lze slouit
na dva trojhelnky spojen podstavou. Tmto slouenm mohou vzniknout
dal diamanty. */
struct SbROAMMergeQueueDiamond
{
  public:
    /* Metody. */
    /** Konstruktor.
    Vytvo instanci diamantu pro prioritn frontu diamant na spojen
    z trojhelnk \p first, \p second, \p third a \p fourth a nastav
    prioritu tohoto diamantu na hodnotu \p priority.
    \param first Prvn trojhelnk diamantu.
    \param second Druh trojhelnk diamantu.
    \param third Tet trojhelnk diamantu.
    \param fourth tvrt trojhelnk diamantu.
    \param priority Priorita diamantu ve front diamant. */
    SbROAMMergeQueueDiamond(SbROAMSplitQueueTriangle * first = NULL,
      SbROAMSplitQueueTriangle * second = NULL,
      SbROAMSplitQueueTriangle * third = NULL,
      SbROAMSplitQueueTriangle * fourth = NULL,
      float priority = PRIORITY_MAX);
    /** Zskn priority diamantu ve front.
    Vrt hodnotu priority diamantu ve front diamant na spojen.
    \return Priorita diamantu ve front diamant na spojen. */
    inline float getPriority() const;
    /** Nastaven priority diamantu ve front.
    Nastav hodnotu priority diamantu ve front \p merge_queue na novou
    hodnotu \p priority.
    \param merge_queue Fronta, v n se diamant nachz.
    \param priority Nov priorita diamantu. */
    inline void setPriority(SbROAMMergeQueue * merge_queue, const float
      priority);
    /** Nastaven priority diamantu.
    Nastav hodnotu priority diamantu na novou hodnotu \p priority, ale
    neaktualizuje jeho pozici ve front.
    \param priority Nov priorita diamantu. */
    inline void setPriority(const float priority);
    /* Datove polozky. */
    /// Prvn trojhelnk diamantu.
    SbROAMSplitQueueTriangle * first;
    /// Druh trojhelnk diamantu.
    SbROAMSplitQueueTriangle * second;
    /// Tet trojhelnk diamantu.
    SbROAMSplitQueueTriangle * third;
    /// tvrt trojhelnk diamantu.
    SbROAMSplitQueueTriangle * fourth;
  private:
    /// Priorita diamantu ve front diamant na spojen.
    float priority;
    /// Index diamantu ve front diamant na spojen.
    int index;
  /** Prioritn fronta diamant na spojen me pistupovat k privtnm
  atributm diamantu. */
  friend class SbROAMMergeQueue;
};

#include <roam/SbROAMSplitQueue.h>
#include <roam/SbROAMMergeQueue.h>

/******************************************************************************
* SbROAMSplitQueueTriangle - public
******************************************************************************/

inline float SbROAMSplitQueueTriangle::getPriority() const
{
  return -priority; // hledame maximum, takze opacna hodnota priority
}

inline void SbROAMSplitQueueTriangle::setPriority(SbROAMSplitQueue *
  split_queue, const float _priority)
{
  /* Nastaveni nove priority a opraveni fronty. */
  priority = -_priority;
  split_queue->newWeight(this, index);
}

inline void SbROAMSplitQueueTriangle::setPriority(const float _priority)
{
  /* Nastaveni nove priority. */
  priority = -_priority;
}

/******************************************************************************
* SbROAMMergeQueueDiamond - public
******************************************************************************/

inline float SbROAMMergeQueueDiamond::getPriority() const
{
  return priority; // vraceni priority
}

inline void SbROAMMergeQueueDiamond::setPriority(SbROAMMergeQueue * merge_queue,
  const float _priority)
{
  /* Nastaveni nove priority a opraveni fronty. */
  priority = _priority;
  merge_queue->newWeight(this, index);
}

inline void SbROAMMergeQueueDiamond::setPriority(const float _priority)
{
  /* Nastaveni nove priority. */
  priority = _priority;
}

#endif
