#ifndef SO_SIMPLE_ROAM_TERRAIN_H
#define SO_SIMPLE_ROAM_TERRAIN_H

///////////////////////////////////////////////////////////////////////////////
//  SoTerrain
///////////////////////////////////////////////////////////////////////////////
/// Tern vykreslovan algoritmem ROAM.
/// \file SoSimpleROAMTerrain.h
/// \author Radek Barto - xbarto33
/// \date 25.08.2005
///
/// Uzel grafu scny reprezentujc tern vykreslovan algoritmem ROAM. Pro
/// pouit je teba uzlu pedadit uzel s koordinty typu \p SoCordinate3
/// obsahujc vkovou mapu ternu a nastavit jej rozmry. Velikost vkov
/// mapy mus bt tvercov o velikosti strany 2^n + 1, kde n je cel kladn
/// slo.
///////////////////////////////////////////////////////////////////////////////
//
// THIS SOFTWARE IS NOT COPYRIGHTED
//
// This source code is offered for use in the public domain.
//  You may use, modify or distribute it freely.
//
// This source code is distributed in the hope that it will be useful but
// WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
// DISCLAIMED.  This includes but is not limited to warranties of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// If you find the source code useful, authors will kindly welcome
// if you give them credit and keep their names with their source code.
//
///////////////////////////////////////////////////////////////////////////////

// Coin includy
#include <Inventor/SbBasic.h>
#include <Inventor/SbLinear.h>
#include <Inventor/SoPrimitiveVertex.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoMFFloat.h>
#include <Inventor/nodes/SoShape.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/elements/SoMaterialBindingElement.h>
#include <Inventor/elements/SoCoordinateElement.h>
#include <Inventor/elements/SoTextureCoordinateElement.h>
#include <Inventor/elements/SoNormalElement.h>
#include <Inventor/elements/SoNormalBindingElement.h>
#include <Inventor/elements/SoViewVolumeElement.h>
#include <Inventor/elements/SoViewportRegionElement.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/bundles/SoMaterialBundle.h>
#include <Inventor/bundles/SoTextureCoordinateBundle.h>

// OpenGL includy
#if defined(__WIN32__) || defined(_WIN32)
  #include <windows.h>
#endif
#include <GL/gl.h>

// ostatni includy
#include <iostream>

// lokalni includy
#include <roam/SbROAMPrimitives.h>
#include <roam/SbROAMSplitQueue.h>
#include <roam/SbROAMMergeQueue.h>
#include <profiler/PrProfiler.h>
#include <debug.h>

/** Tern vykreslovan algoritmem ROAM.
Uzel grafu scny reprezentujc tern vykreslovan algoritmem ROAM. Pro pouit
je teba uzlu pedadit uzel s koordinty typu \p SoCordinate3 obsahujc
vkovou mapu ternu a nastavit jej rozmry. Velikost vkov mapy mus bt
tvercov o velikosti strany 2^n + 1, kde n je cel kladn slo. */
class SoSimpleROAMTerrain : public SoShape
{
  SO_NODE_HEADER(SoSimpleROAMTerrain);
  public:
    /* Metody */
    /** Run-time inicializace tdy.
    Tuto metodu je teba zavolat ped vytvoenm jakkoliv instance tdy
    ::SoSimpleROAMTerrain. */
    static void initClass();
    /** Konstruktor.
    Vytvo instanci tdy ::SoSimpleROAMTerrain, inicializuje ob prioritn
    fronty. */
    SoSimpleROAMTerrain();
    /* Pole. */
    /// Velikost (vka i ka) vkov mapy ternu.
    SoSFInt32 mapSize;
    /// Chyba zobrazen v pixelech.
    SoSFInt32 pixelError;
    /// Maximln poet trojhelnk v triangulaci.
    SoSFInt32 triangleCount;
    /// Pznak oezvn ternu mimo zorn pole.
    SoSFBool frustrumCulling;
    /// Pznak "zmrazen" vykreslovn ternu.
    SoSFBool freeze;
  protected:
    /* Metody */
    /** Vykreslen aktuln triangulace.
    Na zklad vstupnch dat ve vkov map, pozici a smru kamery a dalch
    vlastnost v prvnm sputn vybuduje binrn strom trojhelnk, pot -
    a tak v dalch sputnch - rozdluje a spojuje trojhelnky a diamanty v
    prioritnch frontch tak, aby se doshlo minimln triangulace. Nsledn
    tuto triangulaci vykresl.
    \param action Objekt nesouc informace o grafu scny. */
    virtual void GLRender(SoGLRenderAction * action);
    /** Vytvoen vech trojhelnk vkov mapy.
    Ze vstupn vkov mapy vytvo seznam trojhelnk na nejvy rovni
    detail pro ely detekce koliz a podobn.
    \param action Objekt nesouc informace o grafu scny. */
    virtual void generatePrimitives(SoAction * action);
    /** Vpoet ohranien ternu.
    Vypote rozmry a sted kvdru, kter ohraniuje cel tern.
    \param action Objekt nesouc informace o grafu scny.
    \param box Vypoten kvdr ohraniujc tern.
    \param center Vypoten sted kvdru ohraniujcho tern. */
    virtual void computeBBox(SoAction * action, SbBox3f & box,
      SbVec3f & center);
    /** Inicializace binrnho stromu trojhelnk.
    Rekurzivn inicializuje vechny trojhelnky binrnho stromu trojhelnk,
    vypot jejich normly (pro plon nebo vpoet plynulho stnovn).
    \param index Index trojhelnku, kter inicializovat.
    \param triangle_tree Ukazatel na koen binrnho stromu trojhelnk. */
    void initTriangle(SbROAMTriangle * triangle_tree, const int index);
    /** Test na viditelnost trojhelnku.
    Otestuje trojhelnk definovan body \p first, \p second a \p apex proti
    pohledovmu tlesu. Nenachzi-li se dn z bodu v pohledu kamery, vrt
    \p FALSE. Pouv zjednodueny, ale do tech rozmr rozeny
    Cohen-Shuterland algoritmus, a proto me vrtit \p TRUE i kdy je
    trojhelnk mimo pohled kamery.
    \param first Prvn vrchol testovanho trojhelnku.
    \param second Druh vrchol testovanho trojhelnku.
    \param apex Tet vrchol testovanho trojhelnku.
    \return Vrt \p false nen-li trojhelnk urit v pohledu kamery, je-li
    v pohledu nebo nejde-li rozhodnout vrat \p TRUE. */
    inline SbBool isInViewVolume(SbVec3f first, SbVec3f
      second, SbVec3f apex);
    /** Vpoet priority trojhelnku v triangulaci.
    Na zklad pozice a orientace kamery vypote prioritu trojhelnku
    \p triangle v triangulaci.
    \param triangle Trojhelnk, jeho priorita se m spotat.
    \return Vrt vypotenou prioritu trojhelnku. */
    inline float computePriority(SbROAMTriangle * triangle);
    /** Vmna sousednho trojhelnku.
    Zjist, kterm sousedem je trojhelnk \p old_neighbour, a zamn ho za
    nov trojhelnk \p new_neighbour.
    \param triangle Trojhelnk, u kterho vmnu provst.
    \param old_neighbour Pvodn soused trojhelnku.
    \param new_neighbour Nov soused trojhelnku. */
    inline void reconnectNeighbour(SbROAMSplitQueueTriangle * triangle,
      SbROAMSplitQueueTriangle * old_neighbour,
      SbROAMSplitQueueTriangle * new_neighbour);
    /** Zkladn rozdlen trojhelnku.
    Vyjme trojhelnk \p parent z prioritn fronty trojhelnk
    na rozdlen. Nalezne jeho potomky v binrnm stromu trojhelnk,
    naalokuje jejich reprezentace v prioritn front a vlo je do n.
    Propoj oba potomky mezi sebou a s okolnmi trojhelnky krom bzovho
    souseda otce. Aby bylo toto mono provst zptn, vrt oba potomky
    \p left_child a \p right_child. Reprezentaci otce neuvoluje,
    protoe je tak teba.
    \param parent Otec, kter se m rozdlit.
    \param left_child Lev potomek rozdlenho otce.
    \param right_child Prav potomek rozdlenho otce. */
    void simpleSplit(SbROAMSplitQueueTriangle * parent,
      SbROAMSplitQueueTriangle *& left_child,
      SbROAMSplitQueueTriangle *& right_child);
    /** Obecn rozdlen trojhelnku.
    Kompletn rozdl trojhelnk \p parent z prioritn fronty trojhelnk
    na rozdlen. Pro vlastn rozdlen vol ::simpleSplit. Na rozdlen
    trojhelnky napoj vledek rozdlen bzovho souseda, kter me bt
    dle dlen rekurzivn. Rozdlenho otce dealokuje.
    \param parent Otec, kter se m rozdlit.
    \param left_child Lev potomek rozdlenho otce.
    \param right_child Prav potomek rozdlenho otce. */
    void forceSplit(SbROAMSplitQueueTriangle * parent,
      SbROAMSplitQueueTriangle *& left_child,
      SbROAMSplitQueueTriangle *& right_child);
    /** Spojen poloviny diamant.
    Spoj dva trojhelnky \p left_child a \p right_child diamantu do jednoho
    rodiovskho trojhelnku \p parent. Me vytvoit nov diamanty.
    \param left_child Lev trojhelnk z diamatu.
    \param right_child Prav trojhelnk z diamantu.
    \param parent Vsledn rodiovsk trojhelnk. */
    void halfMerge(SbROAMSplitQueueTriangle * left_child,
      SbROAMSplitQueueTriangle * right_child,
      SbROAMSplitQueueTriangle *& parent);
    /** Spojen diamantu.
    Spoj diamant \p diamond z prioritn fronty diamant na spojen, a
    vytvo tak dva trojhelnky z pvodnch ty v prioritn front
    trojhelnk na rozdlen.
    \param diamond Diamant z prioritn fronty na rozdlen. */
    void merge(SbROAMMergeQueueDiamond * diamond);
    /* Callbacky. */
    /** Callback zmny pole \p ::mapSize.
    Pi zmn hodnoty pole \p ::mapSize nastav jeho intern reprezentaci novou
    hodnotu.
    \param instance Ukazatel na instanci tdy \p SoSimpleROAMTerrain.
    \param sensor Senzor, kter callback vyvolal. */
    static void mapSizeChangedCB(void * instance, SoSensor * sensor);
    /** Callback zmny pole \p ::pixelError.
    Pi zmn hodnoty pole \p ::pixelError nastav jeho intern reprezentaci
    novou hodnotu.
    \param instance Ukazatel na instanci tdy \p SoSimpleROAMTerrain.
    \param sensor Senzor, kter callback vyvolal. */
    static void pixelErrorChangedCB(void * instance, SoSensor * sensor);
    /** Callback zmny pole \p ::triangleCount.
    Pi zmn hodnoty pole \p ::triangleCount nastav jeho intern reprezentaci
    novou hodnotu.
    \param instance Ukazatel na instanci tdy \p SoSimpleROAMTerrain.
    \param sensor Senzor, kter callback vyvolal. */
    static void triangleCountChangedCB(void * instance, SoSensor * sensor);
    /** Callback zmny pole \p ::frustrumCulling.
    Pi zmn hodnoty pole \p ::frustrumCulling nastav jeho intern
    reprezentaci novou hodnotu.
    \param instance Ukazatel na instanci tdy \p SoSimpleROAMTerrain.
    \param sensor Senzor, kter callback vyvolal. */
    static void frustrumCullingChangedCB(void * instance, SoSensor * sensor);
    /** Callback zmny pole \p ::freeze.
    Pi zmn hodnoty pole \p ::freeze nastav jeho intern reprezentaci novou
    hodnotu.
    \param instance Ukazatel na instanci tdy \p SoSimpleROAMTerrain.
    \param sensor Senzor, kter callback vyvolal. */
    static void freezeChangedCB(void * instance, SoSensor * sensor);
    /* Zkratky elementu. */
    /// Body vykov mapy.
    const SbVec3f * coords;
    /// Texturov souadnice pro body vkov mapy.
    const SbVec2f * texture_coords;
    /// Normly.
    const SbVec3f * normals;
    /// Pohledov tleso.
    const SbViewVolume * view_volume;
    /// Vykreslovac okno.
    const SbViewportRegion * viewport_region;
    /// Roviny pohledoveho telesa
    SbPlane planes[6];
    /* Datove polozky. */
    /// Binrn strom trojhelnk.
    SbROAMTriangle * triangle_tree;
    /// Velikost binrnho stromu trojhelnk.
    int tree_size;
    /// Poet rovn binrnho stromu trojhelnk.
    int level;
    /// Konstanta udavajc poet pixel na jeden radin zornho pole.
    float lambda;
    /// Fronta trojhelnk na rozdlen.
    SbROAMSplitQueue * split_queue;
    /// Fronta diamant pro spojen.
    SbROAMMergeQueue * merge_queue;
    /// Pznak pouit textury.
    SbBool is_texture;
    /// Pznak pouit norml.
    SbBool is_normals;
    /* Interni pole. */
    /// Velikost (vka i ka) vkov mapy ternu.
    int map_size;
    /// Chyba zobrazen v pixelech.
    int pixel_error;
    /// Maximln poet trojhelnk v triangulaci.
    int triangle_count;
    /// Pznak oezvn ternu mimo zorn pole.
    SbBool is_frustrum_culling;
    /// Pznak "zmrazen" vykreslovn ternu.
    SbBool is_freeze;
    /* Sensory. */
    /// Senzor pole \p ::mapSize.
    SoFieldSensor * map_size_sensor;
    /// Senzor pole \p ::pixelError.
    SoFieldSensor * pixel_error_sensor;
    /// Senzor pole \p ::triangleCount.
    SoFieldSensor * triangle_count_sensor;
    /// Senzor pole \p ::frustrumCulling.
    SoFieldSensor * frustrum_culling_sensor;
    /// Senzor pole \p ::freeze.
    SoFieldSensor * freeze_sensor;
    /* Konstanty. */
    /// Vchoz hodnota pro chybu triangulace v pixelech.
    static const int DEFAULT_PIXEL_ERROR;
    /// Vchoz hodnota pro maximln poet trojhelnk v triangulaci.
    static const int DEFAULT_TRIANGLE_COUNT;
  private:
    /* Metody */
    /** Destruktor.
    Zprivatizovn, aby se zabrnilo ruen uzlu, nebo Inventor uvoluje pam
    ve vlastn reii. */
    virtual ~SoSimpleROAMTerrain();
};

#endif
