#!/bin/env python
# Name:    PrProfiler log file statistics
# Author:  Radek Barton
# E-mail:  blachex@post.cz
# Date:    02. 01. 2005
# Version: 0.2
#
# Changelog:
#
#   0.2:
#     - maximum and minimum index stat added
#     - some cosmetic changes
#
#   0.1
#     - initial release

import getopt, sys, os;
from math import sqrt;

# Prints usage of a script
def help():
  print "Usage: prstat.py -f log_name -a alg_name [-o obj_id]";
  print "";
  print "-f file_name\t(--file)\tColon separated list of files generated by PrProfiler library.";
  print "-a alg_name\t(--algorithm)\tName of a algorithm to show statistics about.";
  print "-p obj_name\t(--object)\tObject ID to show statistics about.";

# Returns a tuple with one record parsed from line.
def getRecord(line):
  line = line.split(";");
  return (int(line[0]), int(line[1]), line[2], line[3], int(line[4]));

# Square function
def sqr(x):
  return x * x;

# Square difference
def sqrdiff(x, y):
  return sqr(x - y);

# Main function :-).
def main():
  # Reading script arguments
  try:
    options, arguments = getopt.getopt(sys.argv[1:], "f:a:o:p",
      ["file=", "algorithm=", "object=", "parseable"])
  except getopt.GetoptError:
    print "Arguments error!"
    help()
    sys.exit(1)

  # Default arguments values
  file_names = []
  alg_name = None
  object_id = None
  parseable = False

  # Setting arguments values
  for option, value in options:
     if option in ("-f", "--file"):
       file_names = value.split(":")
     elif option in ("-a", "--algorithm"):
       alg_name = value
     elif option in ("-o", "--object"):
       object_id = value
     elif option in ("-p", "--parseable"):
       parseable = True

  # Check required arguments
  if (not file_names) or (not alg_name):
    print "Missing required arguments!"
    help()
    sys.exit(1)

  # File states:
  # 0 - Init state.
  # 1 - Algorithm started state.
  state = 0;
  results = [];

  for file_name in file_names:
    # Per line state machine.
    try:
      file = open(file_name, "r")
    except:
      print "Error openning file %s!" % (file_name)
      sys.exit(1)
    for line in file.readlines():
      # Fetch new line.
      record = getRecord(line)
      # Algorithm didn't start yet.
      if state == 0:
        # Is this the right algorithm record?
        if (record[0] == 1) and (record[3] == alg_name) and \
        (not object_id or (record[2] == object_id)):
          # Algorithm started from now.
          state = 1
          start_record = record
      # Algorithm started.
      elif state == 1:
        # Is this the right algorithm record?
        if (record[0] == 0) and (record[3] == alg_name) and \
        (not object_id or (record[2] == object_id)):
          # Algorithm stopped.
          state = 0
          results += [record[4] - start_record[4]]
    file.close()

  # Check if any results was found
  if len(results) == 0:
    if object_id:
      print "No records found for algorithm %s within object %s." % (alg_name, object_id)
    else:
      print "No records found for algorithm %s." % alg_name
    sys.exit(0)

  # Count statistics
  file_count = len(file_names)
  count = len(results)
  average_count = count / file_count
  suma = sum(results)
  average_suma = suma / file_count
  average = suma / count
  maximum = max(results)
  minimum = min(results)
  max_index = results.index(maximum)
  min_index = results.index(minimum)
  variance = sum(map(sqrdiff, results, [average] * count)) / count
  deviation = int(sqrt(variance) + 0.5)

  # Print results
  if parseable:
    print "%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s" % (alg_name, file_count,
      count, average_count, suma, average_suma, average, maximum, max_index,
      minimum, min_index, variance, deviation)
  else:
    print "Statistics for", alg_name, "algorithm:"
    print "";
    print "Files processed:\t\t", file_count
    print "Appereance count:\t\t", count
    print "Average appereance count:\t", average_count
    print "Complete run time (us):\t\t", suma
    print "Average complete run time (us):\t", average_suma
    print "Average run time (us):\t\t", average
    print "Maximum run time (us):\t\t", maximum
    print "Maximum result index:\t\t", max_index
    print "Minimum run time (us):\t\t", minimum
    print "Minimum result index:\t\t", min_index
    print "Run time variance:\t\t", variance
    print "Run time deviation:\t\t", deviation

if __name__ == "__main__":
  main()
